/*
  rdesktop: A Remote Desktop Protocol client.
  Amiga front-end etc
  Copyright (C) Michal Zukowski 2012-2021
  Copyright (C) Martin Blom 2004-2005

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "rdesktop.h"
#include "config.h"
#include <dos/dos.h>
#include <workbench/startup.h>
#include <workbench/workbench.h>

#include <proto/asl.h>
#include <proto/cybergraphics.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/icon.h>
#include <proto/iffparse.h>
#include <proto/intuition.h>
#include <proto/keymap.h>
#include <proto/layers.h>
#include <proto/utility.h>
#include <proto/wb.h>
#include <ctype.h>		/* isalpha */
#include <stdarg.h>		/* va_list va_start va_end */
#include <unistd.h>		/* read close getuid getgid getpid getppid gethostname */
#include <pwd.h>		/* getpwuid */
#include <sys/signal.h>		/* SIGINT */
#include "amiga_getkeylayout.h"


#ifndef __MORPHOS__
#include <proto/codesets.h>
#include <proto/bsdsocket.h>

struct Library*        CodesetsBase     = NULL;
struct Library*        CyberGfxBase   = NULL;
#endif

extern void rdpsnd_deinit(void);
char *parse_wbargs(char **argv);
extern int amiga_left;
extern int amiga_top;
extern STRPTR amiga_pubscreen_name;
extern ULONG amiga_screen_id;
extern struct DiskObject* amiga_icon;
extern char g_title[64];
extern char g_username[64];
extern char g_hostname[16];
extern char g_keymapname[PATH_MAX];
extern int g_keylayout;
extern int g_requested_session_width;
extern int g_requested_session_height;
extern int g_tcp_port_rdp;
extern int g_server_depth;
extern int g_win_button_size;
extern RD_BOOL g_bitmap_compression;
extern RD_BOOL g_sendmotion;
extern RD_BOOL g_encryption;
extern RD_BOOL g_packet_encryption;
extern RD_BOOL g_desktop_save;
extern RD_BOOL g_polygon_ellipse_orders;
extern RD_BOOL g_fullscreen;
extern RD_BOOL g_clonescreen;
extern RD_BOOL g_grab_keyboard;
extern RD_BOOL g_hide_decorations;
extern RDP_VERSION g_rdp_version;
extern RD_BOOL g_console_session;
extern RD_BOOL g_numlock_sync;
extern RD_BOOL g_owncolmap;
extern RD_BOOL g_ownbackstore;
extern uint32 g_embed_wnd;
extern char g_tls_version[];
extern uint32 g_rdp5_performanceflags;
/* Session Directory redirection */
extern RD_BOOL g_redirect;
extern char g_redirect_server[64];
extern char g_redirect_domain[16];
//extern char g_redirect_password[64];
extern char g_redirect_username[64];
extern char g_redirect_cookie[128];
extern uint32 g_redirect_flags;
extern RD_BOOL g_bitmap_cache_persist_enable;
extern RDPDR_DEVICE g_rdpdr_device[];
extern uint32 g_num_devices;
extern char *g_rdpdr_clientname;

uint32 ext_disc_reason = 0;

extern void show_disconnect_reason(uint16 reason);

const char version[] = "$VER: " PACKAGE_STRING 
                              " (C) 2013-2021 Michal Zukowski, "
                              "(C) 2005-2007 Joerg Strohmayer, "
                              "(C) 2001-2008 Martin Blom, "
                              "(C) 1999-2005 Matthew Chapman et al.";
unsigned long __stack = 1024*100;

extern RD_BOOL g_rdpsnd;
extern int  amiga_audio_unit;


static const char template[] =
"SERVER/A,USER/K,DOMAIN/K,PASSWORD/K,CLIENT/K,CONSOLE/S,FRENCH/S,"
"NOENC=NOENCRYPTION/S,SHELL/K,DIR=DIRECTORY/K,LEFT/K/N,TOP/K/N,"
"W=WIDTH/K/N,H=HEIGHT/K/N,D=DEPTH/K/N,PUBSCREEN/K,TITLE/K,"
"FS=FULLSCREEN/S,CS=CLONESCREEN/S,SCREENMODE/K/N,AUDIO/K/N,REMOTEAUDIO/S,"
"NOMOUSE/S,EXP=EXPERIENCE/K,RDP4/S,KEYMAP/K/N,RDPC=RDPCOMPRESSION/S,"
"PC=PERSISTENTCACHE/S,SM=SLOWMOUSE/S,NETDISK/K,TLSVERSION/K";

static const char wbtemplate[] =
"SERVER/A/K,USER/K,DOMAIN/K,PASSWORD/K,CLIENT/K,CONSOLE/S,FRENCH/S,"
"NOENC=NOENCRYPTION/S,SHELL/K,DIR=DIRECTORY/K,LEFT/K/N,TOP/K/N,"
"W=WIDTH/K/N,H=HEIGHT/K/N,D=DEPTH/K/N,PUBSCREEN/K,TITLE/K,"
"FS=FULLSCREEN/S,CS=CLONESCREEN/S,SCREENMODE/K/N,AUDIO/K/N,REMOTEAUDIO/S,"
"NOMOUSE/S,EXP=EXPERIENCE/K,RDP4/S,KEYMAP/K/N,RDPC=RDPCOMPRESSION/S,"
"PC=PERSISTENTCACHE/S,SM=SLOWMOUSE/S,NETDISK/K,TLSVERSION/K,IGNORE/F";

static LONG  def_size  = 0;
static LONG  def_depth = 0;

static struct
{
    STRPTR a_server;
    STRPTR a_user;
    STRPTR a_domain;
    STRPTR a_password;
    STRPTR a_client;
    ULONG  a_console;
    ULONG  a_french;
    ULONG  a_noenc;
    STRPTR a_shell;
    STRPTR a_dir;
    ULONG* a_left;
    ULONG* a_top;
    LONG*  a_width;
    LONG*  a_height;
    ULONG* a_depth;
    STRPTR a_pubscreen;
    STRPTR a_title;
    ULONG  a_fullscreen;
	ULONG  a_clonescreen;
    ULONG* a_screenmode;
    ULONG* a_audio;
    ULONG  a_remoteaudio;
    ULONG  a_nomouse;
    STRPTR a_experience;
    ULONG  a_rdp4;
    ULONG* a_keymap;
    ULONG  a_rdp_compression;
    ULONG  a_persist_cache;
    ULONG  a_slowmouse;
    STRPTR a_netdisk;
	STRPTR a_tlsversion;
    STRPTR a_ignore;
} a_args =
{
   NULL,
   NULL,
   NULL,
   NULL,
   NULL,
   FALSE,
   FALSE,
   FALSE,
   NULL,
   NULL,
   NULL,
   NULL,
   &def_size,
   &def_size,
   &def_depth,
   NULL,
   NULL,
   FALSE,
   FALSE,
   NULL,
   NULL,
   FALSE,
   FALSE,
   "56K",
   FALSE,
   NULL,
   FALSE,
   FALSE,
   FALSE,
   NULL,
   NULL,
   NULL
};

int a_slowmouse;

static void rdesktop_reset_state(void)
{
  rdp_reset_state();
}

static RD_BOOL read_password(char *password, int size)
{
  RD_BOOL ret = False;  
  char *p;

  fprintf(stderr, "Password: \e[8m");

  if (fgets(password, size, stdin) != NULL)
  {
    ret = True;

    /* strip final newline */
    p = strchr(password, '\n');
    if (p != NULL)
      *p = 0;
  }

  fprintf(stderr, "\e[28m\n");

  return ret;
}

static int 
parse_server_and_port(char *server)
{
  char *p;
  int port = 0;

  p = strchr(server, ':');
  if (p != NULL)
  {
   port = g_tcp_port_rdp = strtol(p + 1, NULL, 10);
    *p = 0;
  }
  return port;
}

enum { UI_INIT          = 1,
       RDPSND_INIT      = 2,
       RDPDR_INIT       = 4,
       RDP_CONNECT      = 8,
       UI_CREATE_WINDOW = 16};

static int startup = 0;

static void cleanup(void)
{
  if (ext_disc_reason >= 2) 
  {
    show_disconnect_reason(ext_disc_reason);
  }

  if( amiga_icon != NULL )
  {
    FreeDiskObject( amiga_icon );
  }

  if (startup & RDPSND_INIT) {
#ifdef WITH_RDPSND
	  
   //  rdpsnd_reset_state();
	  rdpsnd_deinit();
#endif
  }

  if (startup & UI_CREATE_WINDOW) 
  {
    extern int g_pstcache_fd[8];
    int i;

    cache_save_state();
    ui_destroy_window();
    for (i = 0 ; i < 8 ; i++)
    {
      if (g_pstcache_fd[i])
      {
        Close(g_pstcache_fd[i]);
        g_pstcache_fd[i]=0;
      }
    }
  }

  if (startup & RDP_CONNECT) {  
    DEBUG(("Disconnecting...\n"));
    rdp_disconnect();
  }

  if (startup & UI_INIT) {
    ui_deinit();
  }
  
  #ifndef __MORPHOS__
  CloseLibrary( CyberGfxBase );
  CyberGfxBase = NULL;
  CloseLibrary(CodesetsBase);
  CodesetsBase = NULL;
  #endif
}


/* Client program */
int main(int argc, char *argv[])
{
	char server[64];
	char net_disk[1024];
	char fullhostname[64];
	char domain[16];
	char password[64];
	char shell[256];
	char directory[256];
	RD_BOOL prompt_password, deactivated;
	struct passwd *pw;
	uint32 flags;
	int port;
	char *p;

	char pubscreen[64];
	struct RDArgs* rdargs = NULL;
	char* wbargs = NULL;
	int rc = RETURN_OK;

	int run_count = 0;	/* Session Directory support */
	BOOL continue_connect = True;	/* Session Directory support */

	atexit(cleanup);
	flags = RDP_INFO_MOUSE | RDP_INFO_DISABLECTRLALTDEL
		| RDP_INFO_UNICODE | RDP_INFO_MAXIMIZESHELL
		| RDP_INFO_ENABLEWINDOWSKEY;
	prompt_password = False;
	domain[0] = password[0] = shell[0] = directory[0] = 0;
	strcpy(g_keymapname, "en-us");
	g_embed_wnd = 0;
#ifndef __MORPHOS__
	
	if(!(( CodesetsBase = OpenLibrary(CODESETSNAME,CODESETSVER))))
	{
	   error( "Unable to open '%s'.\n", "codeset.library" );	
	   return RETURN_FAIL;
	}
  
	CyberGfxBase = OpenLibrary( "cybergraphics.library", 40 );

	// Don't care if we're unable to open it

	

#else
	pw = getpwuid(getuid());
	if((pw != NULL) && (pw->pw_name != NULL))
#endif
	{
		a_args.a_user = pw->pw_name;
	}

	if(gethostname(fullhostname, sizeof(fullhostname)) != -1)
	{
		p = strchr(fullhostname, '.');
		if(p != NULL)
			*p = 0;

		a_args.a_client = fullhostname;
	}

	g_keylayout = getkeylayout(g_keylayout);

	if(argc == 0)
	{
		wbargs = parse_wbargs(argv);
	}

	if(amiga_icon == NULL)
	{
		BPTR cd = CurrentDir(GetProgramDir());
		char name[ 256 ];

		if(GetProgramName(name, sizeof(name)))
		{
			amiga_icon = GetDiskObjectNew(name[0] ? name : "RDesktop");
		}

		CurrentDir(cd);
	}

	if(0 == argc && wbargs)
	{
		// Fix for ReadArgs: replace all but the last LF with a space
		int length = strlen(wbargs);

		for(int i = 0; i < length - 1; ++i)
		{
			if(wbargs[ i ] == '\n')
			{
				wbargs[ i ] = ' ';
			}
		}

		rdargs = AllocDosObjectTags(DOS_RDARGS, TAG_DONE);

		if(rdargs != NULL)
		{
			rdargs->RDA_Source.CS_Buffer = wbargs;
			rdargs->RDA_Source.CS_Length = length;
			rdargs->RDA_Source.CS_CurChr = 0;
		}
		rdargs = ReadArgs(wbtemplate, (LONG*) &a_args, rdargs);
	}
	else
	{
		rdargs = ReadArgs(template, (LONG*)&a_args, rdargs);
	}

	if(rdargs == NULL)
	{
		char txt[128];

		Fault(IoErr(), NULL, txt, sizeof(txt));
		error("%s\n", txt);
		return RETURN_ERROR;
	}


	if(a_args.a_user == NULL)
	{
		error("Could not determine username.\nYou have to specify it manually.\n");
		rc = RETURN_ERROR;
	}

	if(a_args.a_client == NULL)
	{
		error("Could not determine client name.\nYou have to specify it manually.\n");
		rc = RETURN_ERROR;
	}

	if(a_args.a_server != NULL) STRNCPY(server, a_args.a_server, sizeof(server));
	if(a_args.a_user != NULL)
	{
		STRNCPY(g_username, a_args.a_user, sizeof(g_username));
	}
	if(a_args.a_domain != NULL) STRNCPY(domain, a_args.a_domain, sizeof(domain));
	if(a_args.a_shell != NULL)  STRNCPY(shell, a_args.a_shell, sizeof(shell));
	if(a_args.a_dir != NULL)    STRNCPY(directory, a_args.a_dir, sizeof(directory));
	if(a_args.a_password != NULL)
	{
		STRNCPY(password, a_args.a_password, sizeof(password));
		flags |= RDP_INFO_AUTOLOGON;;
	}
	if(a_args.a_client != NULL) STRNCPY(g_hostname, a_args.a_client, sizeof(g_hostname));

	if(a_args.a_pubscreen != NULL)
	{
		STRNCPY(pubscreen, a_args.a_pubscreen, sizeof(pubscreen));
		amiga_pubscreen_name = pubscreen;
	}

	if(a_args.a_screenmode != NULL)
	{
		amiga_screen_id = *a_args.a_screenmode;
		a_args.a_fullscreen = TRUE;
	}

	g_fullscreen      = a_args.a_fullscreen;
	g_clonescreen     = a_args.a_clonescreen;
	if(a_args.a_left != NULL) amiga_left = *a_args.a_left;
	if(a_args.a_top != NULL)  amiga_top  = *a_args.a_top;

	if(a_args.a_fullscreen && (a_args.a_left != NULL || a_args.a_top != NULL || a_args.a_pubscreen != NULL))
	{
		warning("The LEFT, TOP and PUBSCREEN arguments are ignored in fullscreen mode.\n");
	}

	g_requested_session_width             = *a_args.a_width;
	g_requested_session_height            = *a_args.a_height;

	g_encryption        = ! a_args.a_french;
	g_packet_encryption = ! a_args.a_noenc;
	g_sendmotion        = ! a_args.a_nomouse;
#ifdef WITH_RDPSND
	if(a_args.a_audio != NULL) amiga_audio_unit = *a_args.a_audio;
#endif
	if(a_args.a_remoteaudio)
	{
		flags |= RDP_INFO_REMOTE_CONSOLE_AUDIO;

		if(a_args.a_audio != NULL)
		{
			warning("The AUDIO argument is ignored when REMOTEAUDIO is used.\n");
		}
	}
#ifdef WITH_RDPSND
	else
	{
		g_rdpsnd = True;
	}
#endif

	if(a_args.a_title != NULL) STRNCPY(g_title, a_args.a_title, sizeof(g_title));

	g_server_depth = *a_args.a_depth;

	if(g_server_depth < 0 ||
	        (g_server_depth > 8 && g_server_depth != 16 && g_server_depth != 15
	         && g_server_depth != 24))
	{
		error("invalid server colour depth.\n");
		rc = RETURN_ERROR;
	}

	if(a_args.a_experience != NULL)
	{
		toupper_str(a_args.a_experience);

		if(strcmp("28K8", a_args.a_experience) == 0)
		{
			g_rdp5_performanceflags =
			    PERF_DISABLE_WALLPAPER | PERF_DISABLE_FULLWINDOWDRAG |
			    PERF_DISABLE_MENUANIMATIONS | PERF_DISABLE_THEMING;
		}
		else if(strcmp("56K", a_args.a_experience) == 0)
		{
			g_rdp5_performanceflags =
			    PERF_DISABLE_WALLPAPER | PERF_DISABLE_FULLWINDOWDRAG |
			    PERF_DISABLE_MENUANIMATIONS;
		}
		else if(strcmp("DSL", a_args.a_experience) == 0)
		{
			g_rdp5_performanceflags = PERF_DISABLE_WALLPAPER;
		}
		else if(strcmp("LAN", a_args.a_experience) == 0)
		{
			g_rdp5_performanceflags = PERF_ENABLE_FONT_SMOOTHING;
		}
		else
		{
			error("Illegal EXPERIENCE value: %s\n", a_args.a_experience);
			rc = RETURN_ERROR;
		}
	}

	if(a_args.a_netdisk)
	{
		strcpy(net_disk, "netdisk=");
		strcat(net_disk, a_args.a_netdisk);

		g_num_devices = disk_enum_devices(&g_num_devices, net_disk);
	}
	
	
	if(a_args.a_tlsversion)
	{
		strncpy(g_tls_version, a_args.a_tlsversion, 4);
	}
	
	g_console_session = a_args.a_console;
	g_rdp_version = (a_args.a_rdp4)?RDP_V4:RDP_V5;

	if(a_args.a_keymap != NULL) g_keylayout = *a_args.a_keymap;

	if(a_args.a_rdp_compression) flags |= (RDP_INFO_COMPRESSION | RDP_INFO_COMPRESSION2);
	g_bitmap_cache_persist_enable = 0;
	g_bitmap_cache_persist_enable = a_args.a_persist_cache;
	if(g_sendmotion && a_args.a_slowmouse)
	{
		a_slowmouse = True;
		g_sendmotion = False;
	}

	if(server[0] == 0)
	{
		error("No server specified\n");
		rc = RETURN_ERROR;
	}

	port = parse_server_and_port(server);

	if(strcmp(password, "-") == 0)
	{
		read_password(password, sizeof(password));
	}

	if(g_title[0] == 0)
	{
		if(port)
			snprintf(g_title, sizeof(g_title) - 1, "RDesktop - %s:%d", server, port);
		else snprintf(g_title, sizeof(g_title) - 1,"RDesktop - %s", server);
		g_title[sizeof(g_title) - 1] = '\0';
	}

	FreeArgs(rdargs);
	FreeVec(wbargs);

	// We don't support this at the moment
	g_polygon_ellipse_orders = False;

#define PRINTS( s ) printf( "%s: %s\n", #s, s );
#define PRINTI( i ) //printf( "%s: %d\n", #i, i );

/*	PRINTI(amiga_left);
	PRINTI(amiga_top);
	PRINTI(amiga_screen_id);
	PRINTS(amiga_pubscreen_name);

	PRINTS(g_title);
	PRINTS(g_username);
	PRINTS(password);
	PRINTS(domain);
	PRINTS(g_hostname);
	PRINTS(g_keymapname);
	PRINTI(g_keylayout);
	PRINTI(g_requested_session_width);
	PRINTI(g_requested_session_height);
	PRINTI(tcp_port_rdp);
	PRINTI(g_server_depth);
	PRINTI(g_win_button_size);
	PRINTI(g_bitmap_compression);
	PRINTI(g_sendmotion);
	PRINTI(g_encryption);
	PRINTI(g_packet_encryption);
	PRINTI(g_desktop_save);
	PRINTI(g_fullscreen);
	PRINTI(g_grab_keyboard);
	PRINTI(g_hide_decorations);
	PRINTI(g_rdp_version);
	PRINTI(g_console_session);
	PRINTI(g_numlock_sync);
	PRINTI(g_owncolmap);
	PRINTI(g_ownbackstore);
	PRINTI(g_embed_wnd);
	PRINTI(g_rdp5_performanceflags);
*/

	if(rc != RETURN_OK)
	{
		return rc;
	}

	if(!ui_init())
		return RETURN_FAIL;

	startup |= UI_INIT;

#ifdef WITH_RDPSND
	if(g_rdpsnd && rdpsnd_init(NULL))
	{
		startup |= RDPSND_INIT;	
	}
#endif
	rdpdr_init();
	startup |= RDPDR_INIT;

	while(1)	/* add support for Session Directory; only reconnect once */
	{
		rdesktop_reset_state();

		if(g_redirect)
		{
			STRNCPY(domain, g_redirect_domain, sizeof(domain));
			STRNCPY(g_username, g_redirect_username, sizeof(g_username));
		//	STRNCPY(password, g_redirect_password, sizeof(password));
			STRNCPY(server, g_redirect_server, sizeof(server));
			flags |= RDP_INFO_AUTOLOGON;;
		}


		if(!rdp_connect(server, flags, domain, password, shell, directory, g_redirect))
			return EX_PROTOCOL;

		startup |= RDP_CONNECT;
		/* By setting encryption to False here, we have an encrypted login
		   packet but unencrypted transfer of other packets */
		if(!g_packet_encryption)
			g_encryption = False;

		DEBUG(("Connection successful.\n"));
		memset(password, 0, sizeof(password));

		if(!g_redirect)
		{
			if(!ui_create_window(g_requested_session_width, g_requested_session_height))
			{
				return EX_OSERR;
			}
			else
			{
				startup |= UI_CREATE_WINDOW;
			}
		}
		g_redirect = False;
		tcp_run_ui(True);
		rdp_main_loop(&deactivated, &ext_disc_reason);
		tcp_run_ui(False);
		rdp_disconnect();    
		startup &= ~RDP_CONNECT;


		/*if (g_redirect)
			continue;
		*/
#ifndef ENABLE_AMIGA
		if(g_pending_resize)
		{
			/* If we have a pending resize, reconnect using the new size, rather than exit */
			g_pending_resize = False;
			continue;
		}
#endif
		break;
	}

	if(deactivated)
		return RETURN_OK;
	else if(ext_disc_reason == ERRINFO_RPC_INITIATED_DISCONNECT
	        || ext_disc_reason == ERRINFO_RPC_INITIATED_LOGOFF)
		return RETURN_OK;
	else
		return RETURN_WARN;
}

void amiga_req(char* prefix, char* txt);

static void showmsg(char* prefix, char* format, va_list ap)
{
  char txt[256];

  vsnprintf(txt, sizeof(txt), format, ap);
  va_end(ap);

  {	
    int lf = strlen( txt ) - 1;

    if( lf >= 0 &&  txt[ lf ] == '\n' )
    {
      txt[ lf ] = '\0';
    }
    
    amiga_req( prefix, txt );
  }
}

/* report an error */
void error(char *format, ...)
{
  va_list ap;
  va_start(ap, format);

  showmsg( "ERROR", format, ap );
  va_end(ap);
}

/* report a warning */
void warning(char *format, ...)
{
  va_list ap;
  va_start(ap, format);

  showmsg( "WARNING", format, ap );
  va_end(ap);
}

/* report an unimplemented protocol feature */
void unimpl(char *format, ...)
{
  va_list ap;
  va_start(ap, format);

  showmsg( "NOT IMPLEMENTED", format, ap );
  va_end(ap);
}
